"""Common midnite modules."""
from torch import Tensor
from torch.nn import Module


class Flatten(Module):
    """One layer module that flattens its input, except for minibatch dimension."""

    def forward(self, x: Tensor) -> Tensor:
        if len(x.size()) == 1:
            return x
        else:
            return x.view(x.size(0), -1)



from ..uncertainty import functional
from .modules import Ensemble
from .modules import EnsembleBegin
from .modules import EnsembleLayer
from .modules import MeanEnsemble
from .modules import MutualInformationUncertainty
from .modules import PredictionAndUncertainties
from .modules import PredictiveEntropy
from .modules import StochasticDropouts
from .modules import StochasticModule
from .modules import VariationRatio

__all__ = [
    "StochasticModule",
    "StochasticDropouts",
    "Ensemble",
    "EnsembleBegin",
    "MeanEnsemble",
    "PredictiveEntropy",
    "MutualInformationUncertainty",
    "VariationRatio",
    "PredictionAndUncertainties",
    "EnsembleLayer",
    "functional",
]



from .interface import Activation
from .interface import Attribution
from .interface import LayerSplit
from .interface import NeuronSelector
from .interface import OutputRegularization
from .interface import TransformSequence
from .interface import TransformStep
from .methods import Backpropagation
from .methods import GradAM
from .methods import GuidedBackpropagation
from .methods import Occlusion
from .methods import PixelActivation
from .methods import TVRegularization
from .methods import WeightDecay
from .splits import ChannelSplit
from .splits import Identity
from .splits import NeuronSplit
from .splits import SimpleSelector
from .splits import SpatialSplit
from .splits import SplitSelector
from .transforms import BilateralTransform
from .transforms import BlurTransform
from .transforms import RandomTransform
from .transforms import ResizeTransform

__all__ = [
    "LayerSplit",
    "NeuronSelector",
    "Attribution",
    "Activation",
    "OutputRegularization",
    "TransformStep",
    "TransformSequence",
    "Identity",
    "NeuronSplit",
    "SpatialSplit",
    "ChannelSplit",
    "SimpleSelector",
    "SplitSelector",
    "TVRegularization",
    "WeightDecay",
    "PixelActivation",
    "Backpropagation",
    "GuidedBackpropagation",
    "GradAM",
    "BlurTransform",
    "ResizeTransform",
    "RandomTransform",
    "BilateralTransform",
    "Occlusion",
]



"""Different image transformations, used for regularization in visualization."""
import random as rand

import cv2
import numpy as np
from numpy import ndarray

from ..base.interface import TransformStep


class BlurTransform(TransformStep):
    """Blur transformation step."""

    def __init__(self, blur_size: int = 5):
        """
        Args:
            blur_size: how many pixels to blur

        """
        self.blur_size = blur_size

    def transform(self, img: ndarray) -> ndarray:
        return cv2.blur(img, (self.blur_size, self.blur_size))


class ResizeTransform(TransformStep):
    """Resize transformation step.

    If you use this in an iterative optimization, make sure to start with a smaller
    init_size in order not to run out of GPU memory!

    """

    def __init__(self, scale_fac: float = 1.2):
        """
        Args:
            scale_fac: resizing factor of each step. Keep in mind that large factors
             and/or many iterations will create huge memory requirements!

        """
        self.scale_fac = scale_fac

    def transform(self, img: ndarray) -> ndarray:
        return cv2.resize(
            img,
            (0, 0),
            fx=self.scale_fac,
            fy=self.scale_fac,
            interpolation=cv2.INTER_CUBIC,
        )


class RandomTransform(TransformStep):
    """Applies translation, rotation, and scale in a random direction"""

    def __init__(
        self, pixel_shift: int = 1, rot_deg: float = 0.3, scale_fac: float = 0.1
    ):
        """
        Args:
            pixel_shift: maximum amount of pixels to shift
            rot_deg: maximum degree of rotation
            scale_fac: maximum relative scale factor to resize

        """
        self.px = pixel_shift
        self.deg = rot_deg
        self.fac = scale_fac

    def transform(self, img: ndarray) -> ndarray:
        shift = np.float32(
            [
                [1, 0, rand.randint(-self.px, self.px)],
                [0, 1, rand.randint(-self.px, self.px)],
            ]
        )
        rot = cv2.getRotationMatrix2D(
            (img.shape[0] / 2, img.shape[1] / 2), rand.uniform(-self.deg, self.deg), 1
        )
        scale = 1 + rand.uniform(-self.fac, self.fac)

        img = cv2.warpAffine(img, shift, (0, 0))
        img = cv2.warpAffine(img, rot, (0, 0))
        return cv2.resize(
            img, (0, 0), fx=scale, fy=scale, interpolation=cv2.INTER_CUBIC
        )


class BilateralTransform(TransformStep):
    """Bilateral filter step"""

    def __init__(
        self, diameter: int = 9, color_tolerance: int = 75, dist_tolerance: int = 75
    ):
        """
        Args:
            diameter: diameter of pixel neighbourhood for the filter
            color_tolerance: larger means that more distant colors are mixed together
            dist_tolerance: larger means that pixels further away are mixed together

        """
        self.diam = diameter
        self.color_sigma = color_tolerance
        self.space_sigma = dist_tolerance

    def transform(self, img: ndarray) -> ndarray:
        return cv2.bilateralFilter(img, self.diam, self.color_sigma, self.space_sigma)



"""Some common functionality for visualization methods."""
from torch import Tensor

from midnite.visualization.base import NeuronSelector


def calculate_single_mean(out, select: NeuronSelector) -> Tensor:
    """Calculates the selected mean for a minibatch with one element, retaining grads.

    Args:
        out: network output of shape (1, c, h, w)
        select: selector

    Returns:
        the calculated mean as scalar tensor

    """
    out = out.squeeze(dim=0)
    mask = select.get_mask(list(out.size()))
    return (out * mask).sum() / mask.sum()



"""Concrete implementations of gradient attribution mapping methods."""
from torch import Tensor
from torch.nn import functional
from torch.nn import Module

import midnite
from ...base.interface import Attribution
from ...base.interface import LayerSplit
from ...base.interface import NeuronSelector
from .backpropagation import Backpropagation


class GradAM(Attribution):
    """Gradient attribution mapping.

    Scales the propagated input with the gradient of its inverse split.
    Interpretation: How much does neuron x of layer X (bottom layer selection)
    contribute to neuron y of layer Y (top layer selection)?

    """

    def __init__(
        self,
        inspection_net: Module,
        top_layer_selector: NeuronSelector,
        base_net: Module,
        bottom_layer_split: LayerSplit,
    ):
        """

        Args:
            inspection_net: part of the network from selected until output layer
            top_layer_selector: specifies selected neurons in the output layer
            base_net: part of the network from first layer up to selected
            bottom_layer_split: specifies the split of interest in the selected layer

        """
        super().__init__([], inspection_net, top_layer_selector, bottom_layer_split)
        self.base_net = base_net
        self.backpropagator = Backpropagation(
            inspection_net, self.top_layer_selector, self.bottom_layer_split.invert()
        )

    def visualize(self, input_: Tensor) -> Tensor:
        # Prepare layers/input
        self.base_net.to(midnite.get_device())
        input_ = input_.clone().detach().to(midnite.get_device())

        # forward pass through base layers
        intermediate = self.base_net(input_)

        # retrieve the alpha weight
        intermediate_back = self.backpropagator.visualize(intermediate)
        # get rid of batch dimension
        alpha = intermediate_back.detach().squeeze(dim=0)
        intermediate = intermediate.detach().squeeze(dim=0)
        # fill dimensions s.t. each split has dim (c, h, w)
        alpha = self.bottom_layer_split.invert().fill_dimensions(
            alpha, len(intermediate.size())
        )

        result = self.bottom_layer_split.get_mean(intermediate.mul_(alpha))

        return functional.relu(result)



"""Compound methods created from building block, implementing common use cases."""
from typing import List
from typing import Tuple

import torch
from torch import Tensor
from torch.nn import functional
from torch.nn import Softmax
from torch.nn.modules import Module
from torch.nn.modules import Sequential

import midnite
from ..uncertainty import EnsembleBegin
from ..uncertainty import EnsembleLayer
from ..uncertainty import MutualInformationUncertainty
from ..uncertainty import StochasticModule
from .base import BilateralTransform
from .base import GradAM
from .base import GuidedBackpropagation
from .base import Identity
from .base import NeuronSelector
from .base import NeuronSplit
from .base import Occlusion
from .base import PixelActivation
from .base import RandomTransform
from .base import ResizeTransform
from .base import SimpleSelector
from .base import SpatialSplit
from .base import SplitSelector
from .base import TVRegularization
from .base import WeightDecay


def _prepare_input(img: Tensor) -> Tensor:
    img = img.clone()
    if len(img.size()) == 3:
        img = img.unsqueeze(0)
    if not len(img.size()) == 4 or not img.size(1) == 3:
        raise ValueError(f"Not an image. Size: {img.size()}")
    return img.detach().to(midnite.get_device())


def _top_k_mask(out: Tensor, k: int) -> Tensor:
    mask = torch.zeros_like(out, device=midnite.get_device())
    classes = out.topk(dim=0, k=k)[1]
    for i in range(k):
        mask[classes[i]] = 1
    return mask


def _top_k_selector(net: Module, img: Tensor, k: int) -> NeuronSelector:
    """Creates a top-k classes selector.

    Args:
        net: network
        img: prepared input image
        k: number of classes

    Returns:
        neuron selector for the top k classes

    """
    net = net.to(midnite.get_device())
    out = net(img).squeeze(0)
    mask = _top_k_mask(out, k)

    return SimpleSelector(mask)


def _upscale(img: Tensor, size: Tuple[int, int]) -> Tensor:
    return (
        functional.interpolate(
            img.unsqueeze(dim=0).unsqueeze(dim=0),
            size=size,
            mode="bilinear",
            align_corners=True,
        )
        .squeeze(0)
        .squeeze(0)
    )


def guided_backpropagation(
    layers: List[Module], input_image: Tensor, n_top_classes: int = 3
) -> Tensor:
    """Calculates a simple saliency map of the pixel influence on the top n classes.

    Args:
        layers: the whole model in its layers
        input_image: image tensor of dimensions (c, h, w) or (1, c, h, w)
        n_top_classes: the number of classes to take into account

    Returns:
        a saliency heatmap of dimensions (h, w)

    """
    # Find classes to analyze
    input_image = _prepare_input(input_image)
    net = Sequential(*layers).eval()
    class_selector = _top_k_selector(net, input_image, n_top_classes)

    # Apply guided backpropagation
    backprop = GuidedBackpropagation(layers, class_selector, SpatialSplit())
    return backprop.visualize(input_image)


def gradcam(
    features: Module, classifier: Module, input_image: Tensor, n_top_classes: int = 3
) -> Tensor:
    """Performs GradCAM on an input image.

    For further explanations about GradCam, see https://arxiv.org/abs/1610.02391.

    Args:
        features: the spatial feature part of the model, before classifier
        classifier: the classifier part of the model, after spatial features
        input_image: image tensor of dimensions (c, h, w) or (1, c, h, w)
        n_top_classes: the number of classes to calculate GradCAM for

    Returns:
        a GradCAM heatmap of dimensions (h, w)

    """
    # Get selector for top k classes
    input_image = _prepare_input(input_image)
    class_selector = _top_k_selector(
        Sequential(features.eval(), classifier.eval()), input_image, n_top_classes
    )
    # Apply spatial GradAM on classes
    gradam = GradAM(classifier, class_selector, features, SpatialSplit())
    result = gradam.visualize(input_image)
    return _upscale(result, tuple(input_image.size()[2:]))


def graduam(
    features: StochasticModule,
    classifier: StochasticModule,
    input_image: Tensor,
    num_samples: int = 50,
) -> Tensor:
    """Creates a gradient uncertainty attribution heatmap from the input image.

    Args:
        features: spatial feature part as stochastic module
        classifier: classifier part as stochastic module
        input_image: image tensor of dimensions (c, h, w) or (1, c, h, w)
        num_samples: number of stochastic samples to run

    Returns:
        a GradUAM heatmap of dimensions (h, w)

    """
    input_image = _prepare_input(input_image)

    ensemble_start = EnsembleBegin(num_samples).stochastic_eval()
    feature_ensemble = EnsembleLayer(features).stochastic_eval()
    classifier_ensemble = EnsembleLayer(Sequential(classifier, Softmax(dim=1)))
    classifier_ensemble.stochastic_eval()

    gradam = GradAM(
        Sequential(classifier_ensemble, MutualInformationUncertainty()),
        SplitSelector(Identity(), []),
        Sequential(ensemble_start, feature_ensemble),
        SpatialSplit(),
    )
    result = gradam.visualize(input_image).sum(-1)
    return _upscale(result, tuple(input_image.size()[2:]))


def guided_gradcam(
    feature_layers: List[Module],
    classifier_layers: List[Module],
    input_image: Tensor,
    n_top_classes: int = 3,
) -> Tensor:
    """Performs Guided GradCAM on an input image.

    For further explanations about Guided GradCam, see https://arxiv.org/abs/1610.02391.

    Args:
        feature_layers: the spatial feature layers of the model, before classifier
        classifier_layers: the classifier layers of the model, after spatial features
        input_image: image tensor of dimensions (c, h, w) or (1, c, h, w)
        n_top_classes: the number of classes to calculate GradCAM for

    Returns:
        a Guided GradCAM heatmap of dimensions (h, w)

    """
    input_image = _prepare_input(input_image)
    # Create scaled up gradcam image
    cam = gradcam(
        Sequential(*feature_layers).eval(),
        Sequential(*classifier_layers).eval(),
        input_image,
        n_top_classes,
    )

    # Create guided backprop image
    guided_backprop = guided_backpropagation(
        feature_layers + classifier_layers, input_image, n_top_classes
    )
    # Multiply
    return cam.mul_(guided_backprop)


def occlusion(net: Module, input_image: Tensor, n_top_classes: int = 3) -> Tensor:
    """Creates a attribution heatmap by occluding parts of the input image.

    Args:
        net: the network to visualize attribution for
        input_image: image tensor of dimensions (c, h, w) or (1, c, h, w)
        n_top_classes: the number of classes to account for

    Returns:
        a occlusion heatmap of dimensions (h, w)

    """
    input_image = _prepare_input(input_image)
    class_selector = _top_k_selector(net.eval(), input_image, n_top_classes)
    # Apply occlusion
    occlusion_ = Occlusion(net, class_selector, SpatialSplit(), [1, 10, 10], [1, 5, 5])
    result = occlusion_.visualize(input_image)
    return _upscale(result, tuple(input_image.size()[2:]))


def class_visualization(net: Module, class_index: int) -> Tensor:
    """Visualizes a class for a classification network.

    Args:
        net: the network to visualize for
        class_index: the index of the class to visualize

    """
    if class_index < 0:
        raise ValueError(f"Invalid class: {class_index}")

    img = PixelActivation(
        net.eval(),
        SplitSelector(NeuronSplit(), [class_index]),
        opt_n=500,
        iter_n=20,
        init_size=50,
        transform=RandomTransform(scale_fac=0)
        + BilateralTransform()
        + ResizeTransform(1.1),
        regularization=[TVRegularization(5e1), WeightDecay(1e-9)],
    ).visualize()

    return PixelActivation(
        net,
        SplitSelector(NeuronSplit(), [class_index]),
        opt_n=100,
        iter_n=int(50),
        transform=RandomTransform() + BilateralTransform(),
        regularization=[TVRegularization(), WeightDecay()],
    ).visualize(img)



"""Acquisition functions and helpers.

Definitions from https://arxiv.org/pdf/1703.02910.pdf.
See Also https://arxiv.org/pdf/1802.10501.pdf, Appendix C for derivations
of MC approximations.

"""
import logging

import torch
from torch import Tensor

log = logging.getLogger(__name__)


def mean_prediction(input_: Tensor) -> Tensor:
    """Gives the mean over sampled predictions.

    Approximation: p(y|x,D) = 1/T sum_t p(y|x,w_t)

    Args:
        input_: concatenated predictions for K classes and T samples (minibatch size N),
         of shape (N, K, ..., T)

    Returns:
        the mean prediction for all samples, i.e. p(y|x,D), of shape (N, K, ...)

    """
    # Check if all equal predictions, as .mean would introduce numerical error
    if input_.unique(dim=-1).size(-1) == 1:
        return input_.select(dim=-1, index=-1).clone()

    return torch.mean(input_, dim=(-1,))


def predictive_entropy(input_: Tensor, log_clamp=1e-40, inplace=False) -> Tensor:
    """Calculates the predictive entropy over the samples in the input.

    Approximation: H[y|x,D] = - sum_y p(y|x,D) * log p(y|x,D)

    Args:
        log_clamp: min value for prediction mean, to avoid NaN errors.
         Default value: 1e-40
        input_: concatenated predictions for K classes and T samples (minibatch size N),
         of shape (N, K, T, ...)
        inplace: whether to perform the operations in-place

    Returns: the predictive entropy per class,
     i.e. H[y|x,D] = - sum_y p(y|x,D) * log p(y|x,D), of shape (N, K, ...)

    """
    if not torch.all(input_ >= 0) or not torch.all(input_ <= 1):
        raise ValueError("Not a probabilistic input")
    # Min clamp necessary in case of log(0)
    _ensemble_mean = mean_prediction(input_)
    if inplace:
        _ensemble_mean.clamp_(log_clamp)
        return _ensemble_mean.mul_(_ensemble_mean.log()).mul_(-1.0)
    else:
        _ensemble_mean = _ensemble_mean.clamp(log_clamp)
        return _ensemble_mean.mul(_ensemble_mean.log()).mul(-1.0)


def mutual_information_uncertainty(
    input_: Tensor, log_clamp=1e-40, inplace=False
) -> Tensor:
    """Calculates the uncertainty based on mutual information.

    Approximation: I[y,w|x,D] = H[y|x,D] - E[sum_y H[y|x,w]]
        = H[y|x,D] + sum_t,y p(y|x,w_t) * log p(y|x,w_t)

    Args:
        log_clamp: min value for input mean, to avoid NaN errors.
         Default value: 1e-40
        input_: concatenated predictions for K classes and T samples (minibatch size N),
         of shape (N, K, T, ...)
        inplace: whether to perform the operations in-place

    Returns: the mutual information uncertainty, i.e. H[y|x,D] - E[sum_y H[y|x,w]],
     of shape (N, K, ...)

    """
    if not torch.all(input_ >= 0) or not torch.all(input_ <= 1):
        raise ValueError("Not a probabilistic input")
    # Min clamp necessary in case of log(0)
    pred_entropy = predictive_entropy(input_, log_clamp, inplace)

    if inplace:
        clamped_input = input_.clamp_(min=log_clamp)
        expected_entropies = clamped_input.mul_(clamped_input.log())
        result = expected_entropies.add_(pred_entropy.unsqueeze_(-1)).mean(-1)
    else:
        clamped_input = input_.clamp(min=log_clamp)
        expected_entropies = clamped_input.mul(clamped_input.log())
        result = expected_entropies.add(pred_entropy.unsqueeze(-1)).mean(-1)
    if torch.all(result == 0):
        log.warning("Predictions are not stochastic")
    return result


def variation_ratio(input_: Tensor, inplace=False) -> Tensor:
    """Calculates the variation ratio over the samples in the input.

    Args:
        input_: concatenated predictions for K classes and T samples (minibatch size N),
         of shape (N, K, ..., T)
        inplace: whether to perform the operations in-place

    Returns: the variation ratio, i.e. 1 - max_y p(y|x,D), of shape (N, K, ...)

    """
    if not torch.all(input_ >= 0) or not torch.all(input_ <= 1):
        raise ValueError("Not a probabilistic input")
    # shape (N, K, ...)
    mean = mean_prediction(input_)

    # shape (N, ...)
    max_mean = torch.max(mean, dim=1)[0]
    if inplace:
        return max_mean.mul_(-1.0).add_(1.0)
    else:
        return max_mean.mul(-1.0).add(1.0)



"""Concrete implementations of occlusion methods."""
from typing import List

import torch
import tqdm
from torch import Tensor
from torch.nn import functional
from torch.nn import Module

import midnite
from ...base.interface import Attribution
from ...base.interface import LayerSplit
from ...base.interface import NeuronSelector


class Occlusion(Attribution):
    """Calculates attribution by occluding parts of the image and comparing outputs"""

    def __init__(
        self,
        net: Module,
        top_layer_selector: NeuronSelector,
        bottom_layer_split: LayerSplit,
        chunk_size: List[int],
        stride: List[int],
        norm=2,
    ):
        """
        Args:
            net: part of the network to analyze
            top_layer_selector: specifies elected neurons in the output layer
            bottom_layer_split: neuron split of the bottom layer (i.e., input image)
            chunk_size: chunk size in number of strides, for each dimension
            stride: number of elements (usually pixels) to go through per step,
             for each dimension
            norm: the norm to be computed
        """
        super().__init__([], net, top_layer_selector, bottom_layer_split)
        if not len(chunk_size) == len(stride):
            raise ValueError("Stride and chunk size must have same dimensionality")
        self.chunk_size = chunk_size
        self.stride = stride
        if norm < 1:
            raise ValueError("Must be valid distance norm")
        self.norm = norm

    @classmethod
    def _smear_matrix(cls, size: int, smear: int) -> Tensor:
        """Creates a smear matrix with given size and smear factor."""
        # Identity
        matrix = torch.eye(size, device=midnite.get_device())
        # Add smear diagonals
        for i in range(1, smear):
            matrix[i:, :-i] += torch.eye(size - i, device=midnite.get_device())
        return matrix.clamp_(max=1)

    def _chunk_matrixes(self, input_size: List[int]) -> List[Tensor]:
        """For every dimension, builds the chunk matrix of chunk + input size."""
        if not len(input_size) == len(self.chunk_size):
            raise ValueError("Chunks size must be given for every input dimension")
        return list(
            map(
                lambda sizes: self._smear_matrix(sizes[0] + sizes[1], sizes[1]),
                zip(input_size, self.chunk_size),
            )
        )

    def _chunk_mask(self, pixel_mask: Tensor, chunk_matrixes: List[Tensor]) -> Tensor:
        """Translates a pixel selector mask into a chunk selector mask."""
        sizes = list(zip(pixel_mask.size(), self.chunk_size))
        mask_size = tuple(map(lambda s: s[0] + s[1], sizes))
        # Create mask with initial pixel mask
        mask = torch.zeros(mask_size, device=midnite.get_device())
        mask[tuple(map(lambda s: slice(None, s), pixel_mask.size()))] = pixel_mask
        # build mask
        for i, matrix in enumerate(chunk_matrixes):
            mask = (matrix @ mask.transpose(1, i)).transpose(1, i)
        # Get relevant mask area
        mask_area = tuple(
            map(lambda s: slice((s[1] - 1) // 2, s[0] + ((s[1] - 1) // 2)), sizes)
        )
        # Normalize indexes that were selected multiple times
        return mask[mask_area].clamp_(max=1)

    @classmethod
    def _remove_chunk(cls, input_: Tensor, chunk_mask: Tensor) -> Tensor:
        # Upsample to input image dimensions (use volumetric to get 3d upsample)
        mask = functional.interpolate(
            chunk_mask.unsqueeze(1), tuple(input_.squeeze(0).size())
        ).squeeze(1)
        # Fill chunk with grey
        return input_.mul_(1 - mask).add_(mask * 0.5)

    def visualize(self, input_: Tensor) -> Tensor:
        # Preparation
        input_ = input_.clone().detach().to(midnite.get_device())
        self.black_box_net.to(midnite.get_device())

        # Make reference prediction
        pred = self.black_box_net(input_).squeeze(0)
        # Store top layer mask for the future
        pred_mask = self.top_layer_selector.get_mask(list(pred.size()))
        pred_mask.div_(pred_mask.sum())
        pred.mul_(pred_mask)

        # Dimensions of subsampled image
        subsample_dims = list(
            map(lambda t: t[0] // t[1], zip(input_.squeeze(0).size(), self.stride))
        )

        # Calculate chunk matrixes once and store
        chunk_matrixes = self._chunk_matrixes(subsample_dims)

        # For every pixel in subsampled image, measure difference to true prediction
        result_img = torch.zeros(*subsample_dims, device=midnite.get_device())
        for pixel in tqdm.tqdm(self.bottom_layer_split.get_split(subsample_dims)):
            mask = self._chunk_mask(pixel, chunk_matrixes)
            img = self._remove_chunk(input_.clone(), mask.unsqueeze(0))
            out = self.black_box_net(img).squeeze(0).mul_(pred_mask)
            result_img.add_(mask.squeeze(0).mul_(out.dist(pred, self.norm).item()))

        return self.bottom_layer_split.get_mean(result_img)



"""Custom modules for MC dropout ensembles and uncertainty."""
import logging
from abc import ABC
from abc import abstractmethod
from typing import Tuple
from typing import Union

import torch
import tqdm
from torch import Tensor
from torch.nn import AlphaDropout
from torch.nn import Dropout
from torch.nn import Dropout2d
from torch.nn import Dropout3d
from torch.nn import FeatureAlphaDropout
from torch.nn import Module
from torch.nn import Sequential
from tqdm import trange

import midnite
from . import functional as func

log = logging.getLogger(__name__)


def _is_dropout(layer: Module) -> bool:
    return (
        isinstance(layer, Dropout)
        or isinstance(layer, Dropout2d)
        or isinstance(layer, Dropout3d)
        or isinstance(layer, AlphaDropout)
        or isinstance(layer, FeatureAlphaDropout)
    )


class InnerForwardMixin:
    """Mixin for simple forward passes."""

    def forward(self, input_):
        input_.to(midnite.get_device())
        for module in self.children():
            input_ = module(input_)
        return input_


class StochasticModule(Module, ABC):
    """Interface for stochastic models.

    Stochastic models can make different predictions for the same input, e.g. using an
    active dropout layer at prediction time.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stochastic = False

    def train(self, mode=True):
        super().train(mode)
        self.stochastic = False

        # All recursive submodules, including self
        for module in self.modules():
            if module is not self and isinstance(module, StochasticModule):
                module.stochastic = False
        return self

    def stochastic_eval(self):
        """Sets the model to stochastic evaluation mode.

        Override this to set your module to stochastic mode.

        Returns: self for fluent interface

        """
        if self.training:
            self.eval()

        self.stochastic = True

        # All recursive submodules, including self
        for module in self.modules():
            if module is not self and isinstance(module, StochasticModule):
                module.stochastic_eval()
        return self


class StochasticDropouts(InnerForwardMixin, StochasticModule):
    """Modifies all dropout layers to be in train mode durching stochastic_eval time."""

    def __init__(self, inner: Module):
        """
        Args:
            inner: inner module

        """
        super().__init__()
        self.add_module("inner", inner)

    def stochastic_eval(self):
        super().stochastic_eval()

        for module in self.modules():
            if module is not self and _is_dropout(module):
                module.training = True

        return self


class EnsembleBegin(StochasticModule):
    def __init__(self, num_passes: int = 20):
        """
        Args:
            num_passes: number of stochastic passes in the ensemble

        """
        super().__init__()
        if num_passes < 2:
            raise ValueError("At least two samples are necessary")

        if num_passes < 20:
            log.warning("Using low number of samples may give greatly skewed results")

        self.num_passes = num_passes

    def forward(self, input_: Tensor) -> Tensor:
        input_.to(midnite.get_device())

        # In eval mode, stack ensemble predictions
        if self.stochastic:
            if input_.requires_grad:
                log.warning(
                    "Gradients enabled for ensemble predictions requires large "
                    "amounts of memory"
                )

            return torch.stack((input_,) * self.num_passes, len(input_.size()))
        else:
            return input_


class EnsembleLayer(InnerForwardMixin, StochasticModule):
    """Module for a single layer in ensemble mode.

    May be used between EnsembleBegin and an acquisition function, e.g. PredictiveEntropy.

    """

    def __init__(self, inner: Union[Module, StochasticModule]):
        """
        Args:
            inner: stochastic inner module

        """
        super().__init__()
        self.add_module("inner", inner)

    def forward(self, input_: Tensor) -> Tensor:
        if self.stochastic:
            input_.to(midnite.get_device())
            outs = []
            for i in trange(input_.size(-1)):
                single_input = input_.select(-1, i).clone()
                outs.append(super().forward(single_input))
            return torch.stack(tuple(outs), len(outs[0].size()))
        else:
            return super().forward(input_)


class MeanEnsemble(InnerForwardMixin, StochasticModule):
    """Module for models that want the mean of their ensemble predictions."""

    def __init__(self, inner: Union[Module, StochasticModule], num_passes: int = 20):
        """
        Args:
            inner: stochastic inner module
            num_passes: number of stochastic passes in the ensemble

        """
        super().__init__()
        self.add_module("inner", inner)

        if num_passes < 2:
            raise ValueError("At least two samples are necessary")

        if num_passes < 20:
            log.warning("Using low number of samples may give greatly skewed results")

        self.num_passes = num_passes

    def forward(self, input_: Tensor) -> Tensor:
        if self.stochastic:
            with tqdm.trange(
                self.num_passes - 1, total=self.num_passes
            ) as sample_range:
                pred = super().forward(input_)
                sample_range.update()

                # Calc remaining forward pass
                if input_.requires_grad:
                    log.warning(
                        "Gradients enabled for ensemble predictions requires large "
                        "amounts of memeory"
                    )

                    for _ in sample_range:
                        pred = pred.add(super().forward(input_))
                    return pred.div(self.num_passes)
                else:
                    for _ in sample_range:
                        pred.add_(super().forward(input_))
                    return pred.div_(self.num_passes)
        else:
            return super().forward(input_)


class Acquisition(Module, ABC):
    """Abstract base class for acquisition functions."""

    def __init__(self, per_class=False):
        """
        Args:
            per_class: signals whether the uncertainty measure should be outputted
             in total or per class

        """
        super().__init__()
        self.per_class = per_class

    @abstractmethod
    def measure_uncertainty(self, input_: Tensor) -> Tensor:
        """Abstract method to calculate uncertainty

        Args:
            input_: the input tensor

        Returns:
            an uncertainty tensor

        """
        raise NotImplementedError

    def forward(self, input_: Tensor) -> Tensor:
        input_.to(midnite.get_device())
        uncertainty = self.measure_uncertainty(input_)

        if self.per_class:
            return uncertainty
        else:
            return uncertainty.sum(dim=1)


class PredictiveEntropy(Acquisition):
    """Module to calculate the predictive entropy from an ensemble model.

    Predictive entropy (also called max entropy) measures total uncertainty.

    """

    def measure_uncertainty(self, input_: Tensor) -> Tensor:
        return func.predictive_entropy(input_, inplace=not input_.requires_grad)


class MutualInformationUncertainty(Acquisition):
    """Module to calculate an uncertainty based on mutual information.

    Measures epistemic uncertinty. Also called Bayesian Active Learning by Disagreement
     (BALD).

    """

    def measure_uncertainty(self, input_: Tensor) -> Tensor:
        return func.mutual_information_uncertainty(
            input_, inplace=not input_.requires_grad
        )


class VariationRatio(Acquisition):
    """Module to calculate the variation ratio from an ensemble model.

    Gives a percentage for total predictive uncertainty.

    """

    def measure_uncertainty(self, input_: Tensor) -> Tensor:
        return func.variation_ratio(input_, inplace=not input_.requires_grad)


class PredictionAndUncertainties(Acquisition):
    """Module to conveniently calculate sampled mean and uncertainties."""

    def measure_uncertainty(self, input_: Tensor) -> Tuple[Tensor, Tensor]:
        return (
            func.predictive_entropy(input_.clone(), inplace=not input_.requires_grad),
            func.mutual_information_uncertainty(
                input_, inplace=not input_.requires_grad
            ),
        )

    def forward(self, input_: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Returns:
            mean prediction, predictive entropy, mutual information

        """
        input_.to(midnite.get_device())
        pred = input_.mean(dim=(input_.dim() - 1,))
        pe, mi = self.measure_uncertainty(input_)
        if self.per_class:
            return pred, pe, mi
        else:
            return pred, pe.sum(dim=1), mi.sum(dim=1)


class Ensemble(InnerForwardMixin, StochasticModule):
    """Complete ensemble wrapper."""

    def __init__(
        self, begin: EnsembleBegin, acquisition: Acquisition, *layers: EnsembleLayer
    ):
        """
        Args:
            start: the ensemble start
            *layers: the intermediate layers
            acquisition: the final acquisition function
        """
        super().__init__()
        self.add_module("start", begin)
        self.add_module("layers", Sequential(*layers))
        self.add_module("acquisition", acquisition)



from . import base
from .compound_methods import class_visualization
from .compound_methods import gradcam
from .compound_methods import graduam
from .compound_methods import guided_backpropagation
from .compound_methods import guided_gradcam
from .compound_methods import occlusion

__all__ = [
    "base",
    "gradcam",
    "graduam",
    "guided_backpropagation",
    "guided_gradcam",
    "occlusion",
    "class_visualization",
]



from .backpropagation import Backpropagation
from .backpropagation import GuidedBackpropagation
from .gradam import GradAM
from .occlusion import Occlusion
from .pixel_activation import PixelActivation
from .pixel_activation import TVRegularization
from .pixel_activation import WeightDecay

__all__ = [
    "TVRegularization",
    "WeightDecay",
    "PixelActivation",
    "Backpropagation",
    "GuidedBackpropagation",
    "GradAM",
    "Occlusion",
]



from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Optional

import torch
from numpy import ndarray
from torch import Tensor
from torch.nn import Module

import midnite


class LayerSplit(ABC):
    """Abstract base class for splits of a layer.

    A 'split' represents a way to 'slice the cube' of the spatial and
    channel positions of a layer. For more information, see
    https://distill.pub/2018/building-blocks/

    """

    @abstractmethod
    def invert(self) -> "LayerSplit":
        """Returns an inverted split"""
        raise NotImplementedError()

    @abstractmethod
    def get_split(self, size: List[int]) -> List[Tensor]:
        """Returns a split, i.e. all masks.

        Args:
            size: the size of the layer to split, usually as [c, h, w]

        Returns:
            a list of masks containing a mask for each split

        """
        raise NotImplementedError()

    @abstractmethod
    def get_mask(self, index: List[int], size: List[int]) -> Tensor:
        """Returns a single mask for the indexed split.

        Args:
            index: the index of the mask to get from the split
            size: the size of the layer to split, usually as [c, h, w]

        Returns:
            a single split mask

        """
        raise NotImplementedError()

    def get_mean(self, input_: Tensor) -> Tensor:
        """Returns the mean of an input along the split dimension.

        Args:
            input_: Tensor for which the mean should be computed
        Returns:
            a tensor of mean values.

        """
        output = torch.zeros_like(input_, device=midnite.get_device())
        for mask in self.get_split(input_.size()):
            norm = mask.sum()
            output.add_(mask.mul_(input_).div_(norm))
        return output

    def fill_dimensions(self, input_: Tensor, num_dimensions: int = 3) -> Tensor:
        """Fills up the dimensions with unsqueeze() to the given number.

        Args:
            input_: tensor which needs to be unsqueezed
            num_dimensions: number of dimensions to fill up to

        Returns:
            the input with num_dimensions

        """
        dims_to_fill = num_dimensions - len(input_.size())
        if dims_to_fill < 0:
            raise ValueError("Already larger than num_dimensions")
        for i in range(dims_to_fill):
            input_ = input_.unsqueeze(-1)
        return input_


class NeuronSelector(ABC):
    @abstractmethod
    def get_mask(self, size: List[int]) -> Tensor:
        """Get the mask for the specified neurons

        Args:
            size: size of the layer

        Returns:
            a mask of the selected neurons

        """
        raise NotImplementedError()


class Attribution(ABC):
    """Abstract base class for attribution methods.

    Attribution methods aim to extract insights about the effect that neurons in
    different layers have on each other.

    """

    def __init__(
        self,
        white_box_layers: List[Module],
        black_box_net: Optional[Module],
        top_layer_selector: NeuronSelector,
        bottom_layer_split: LayerSplit,
    ):
        """
        Args:
            white_box_layers: list of adjacent layers viewed as white box
            black_box_net: part of the network that is viewed as black box
            top_layer_selector: the target split for analyzing attribution
            bottom_layer_split: split of the selected layer

        """
        if len(white_box_layers) == 0 and black_box_net is None:
            raise ValueError("Must specify at least one layer")
        self.white_box_layers = white_box_layers
        self.black_box_net = black_box_net
        self.top_layer_selector = top_layer_selector
        self.bottom_layer_split = bottom_layer_split

    @abstractmethod
    def visualize(self, input_: Tensor) -> Tensor:
        """Abstract method to call attribution method

        Args:
            input_: the input tensor

        Returns: a tensor showing attribution

        """
        raise NotImplementedError()


class Activation(ABC):
    """Abstract base class for activation methods.

    Activation methods aim to visualize some properties of the activation of a layer or
    a number of layers.

    """

    def __init__(
        self,
        white_box_layers: List[Module],
        black_box_net: Optional[Module],
        top_layer_selector: NeuronSelector,
    ):
        """
        Args:
            white_box_layers: list of adjacent layers viewed as white box
            black_box_net: part of the network that is viewed as black box
            top_layer_selector: the split on which the activation properties are
             extracted

        """
        if len(white_box_layers) == 0 and black_box_net is None:
            raise ValueError("Must specify at least one layer")
        self.white_box_layers = white_box_layers
        self.black_box_net = black_box_net
        self.top_layer_selector = top_layer_selector

    @abstractmethod
    def visualize(self, input_: Optional[Tensor] = None) -> Tensor:
        """Visualizes an activation.

        Args:
            input_: an optional input image, usually a prior. Of shape (height, width,
             channels).

        Returns:
            an activation visualization of shape (h, w, c).

        """
        raise NotImplementedError()


class OutputRegularization(ABC):
    """Base class for regularizations on an output term."""

    def __init__(self, coefficient: float = 0.1):
        """
        Args:
            coefficient: how much regularization to apply

        """
        self.coefficient = coefficient

    @abstractmethod
    def loss(self, out: Tensor) -> Tensor:
        """Calculates the loss for an output.

         Args:
             out: the tensor to calculate the loss for, of shape
              (channels, height,width)

         Returns:
             a tensor representing the loss

         """
        raise NotImplementedError()


class TransformStep(ABC):
    """Abstract base class for an image transformation step"""

    @abstractmethod
    def transform(self, img: ndarray) -> ndarray:
        """Transforms the image.

        Args:
            img: the image to transform

        Returns:
            the transformed image

        """
        raise NotImplementedError()

    def __add__(self, other: "TransformStep"):
        """Method to easily concatenate two transformations.

        Args:
            other: the second transformation to apply

        Returns:
            a TransformSequence containing self and other

        """
        return TransformSequence(self, other)


class TransformSequence(TransformStep):
    """Concatenates a number of transform steps as a sequence."""

    def __init__(self, *steps: TransformStep):
        """
        Args:
            *steps: transformation steps to apply after each other

        """
        self.steps = steps

    def transform(self, img: ndarray) -> ndarray:
        for step in self.steps:
            img = step.transform(img)
        return img

    def __add__(self, other: TransformStep):
        self.steps += (other,)
        return self



"""General interface of the visualization building blocks"""
import itertools
import logging
from typing import List

import torch
from torch import Tensor

import midnite
from ..base.interface import LayerSplit
from ..base.interface import NeuronSelector

log = logging.getLogger(__name__)


class Identity(LayerSplit):
    """Identity selector that implements LayerSplit interface."""

    def fill_dimensions(self, input_: Tensor, num_dimensions: int = 3) -> Tensor:
        if not len(input_.size()) == 0:
            raise ValueError(
                f"Cannot specify element for identity. Got: {input_.size()}"
            )
        return super().fill_dimensions(input_, num_dimensions)

    def invert(self) -> LayerSplit:
        return NeuronSplit()

    def get_split(self, size: List[int]) -> List[Tensor]:
        return [self.get_mask([], size)]

    def get_mask(self, index: List[int], size: List[int]) -> Tensor:
        if len(index) > 0:
            raise ValueError("No index required for identity split")
        return torch.ones(tuple(size), device=midnite.get_device())

    def get_mean(self, input_: Tensor) -> Tensor:
        return input_.mean()


class NeuronSplit(LayerSplit):
    """Split a layer neuron-wise, i.e. per single value."""

    def invert(self) -> LayerSplit:
        return Identity()

    def get_split(self, size: List[int]) -> List[Tensor]:
        indexes = itertools.product(*map(range, size))
        return list(map(lambda idx: self.get_mask(idx, size), indexes))

    def get_mask(self, index: List[int], size: List[int]) -> Tensor:
        mask = torch.zeros(tuple(size), device=midnite.get_device())
        mask[tuple(index)] = 1
        return mask

    def get_mean(self, input_: Tensor) -> Tensor:
        if not len(input_.size()) == 3:
            log.warning(f"Input normally has 3 dimensions. Got: {input_.size()}")
        return input_


class SpatialSplit(LayerSplit):
    """Split a layer by spatial positions."""

    def fill_dimensions(self, input_: Tensor, num_dimensions: int = 3):
        if not len(input_.size()) == 2:
            log.warning(f"Input should have 2 spatial dimensions. Got: {input_.size()}")
        if num_dimensions == 2 and len(input_.size()) == 2:
            return input_
        else:
            return super().fill_dimensions(input_.unsqueeze(0), num_dimensions)

    def invert(self) -> LayerSplit:
        return ChannelSplit()

    def get_split(self, size: List[int]) -> List[Tensor]:
        indexes = itertools.product(*map(range, size[1:]))
        return list(map(lambda idx: self.get_mask(idx, size), indexes))

    def get_mask(self, index: List[int], size: List[int]) -> Tensor:
        if not len(index) == 2:
            raise ValueError("Spatial index need two dimensions. Got: ", index)
        mask = torch.zeros(*size, device=midnite.get_device())
        mask[:, index[0], index[1]] = 1
        return mask

    def get_mean(self, input_: Tensor) -> Tensor:
        if not len(input_.size()) == 3:
            log.warning(f"Input normally has 3 dimensions. Got: {input_.size()}")
        # mean over channel dimension, so that there is one mean value for each spatial
        return input_.mean(0)


class ChannelSplit(LayerSplit):
    """Split a layer by its channels."""

    def fill_dimensions(self, input_: Tensor, num_dimensions: int = 3) -> Tensor:
        if not len(input_.size()) == 1:
            log.warning(f"Input should have 1 channel dimensions. Got: {input_.size()}")
        return super().fill_dimensions(input_, num_dimensions)

    def invert(self) -> LayerSplit:
        return SpatialSplit()

    def get_split(self, size: List[int]) -> List[Tensor]:
        indexes = map(lambda idx: [idx], range(size[0]))
        return list(map(lambda idx: self.get_mask(idx, size), indexes))

    def get_mask(self, index: List[int], size: List[int]) -> Tensor:
        if not len(index) == 1:
            raise ValueError(f"Channel index needs one dimension. Got: {index}")
        mask = torch.zeros(*size, device=midnite.get_device())
        mask[index[0]] = 1
        return mask

    def get_mean(self, input_: Tensor) -> Tensor:
        if not len(input_.size()) == 3:
            log.warning(f"Input normally has 3 dimensions. Got: {input_.size()}")
        # mean over spatials, s.t. there is one mean value for each channel
        return input_.mean(-1).mean(-1)


class GroupSplit(LayerSplit):
    """Splits a layer by a decomposition, usually factorized by (spatials, channels)."""

    def __init__(self, left: Tensor, right: Tensor):
        """

        Args:
            left: the left decomposition matrix. Of shape (..., N) where N is the number
             of groups in the decomposition
            right: the right decomposition matrix. Of shape (N, ...)

        """
        if not left.size(-1) == right.size(0):
            raise ValueError(
                f"Not a valid decomposition. Group dimensions must be exqual. Got: "
                f" {left.size(-1)}, {right.size(0)}"
            )
        self.num_groups = left.size(-1)
        self.left = left
        self.right = right

    def invert(self) -> LayerSplit:
        raise NotImplementedError("Inverting neuron groups is not yet implemented")

    def get_split(self, size: List[int]) -> List[Tensor]:
        indexes = range(self.num_groups)
        return list(map(lambda i: self.get_mask([i], size), indexes))

    def get_mask(self, index: List[int], size: List[int]) -> Tensor:
        if not len(index) == 1:
            raise ValueError("Need to provide exactly one index for a group")
        idx = index[0]
        mask = self.left.select(-1, idx).unsqueeze(-1) @ self.right.select(
            0, idx
        ).unsqueeze(0)
        if not size == list(mask.size()):
            raise ValueError("Requested size does not fit neuron group")
        return mask

    def fill_dimensions(self, input_: Tensor, num_dimensions: int = 3) -> Tensor:
        size = tuple(self.left.size()[:-1] + self.right.size()[1:])
        if not size == input_.size():
            raise ValueError(
                f"Input must of incorrect size. Expected: {size},"
                f" got: {tuple(input_.size())}"
            )
        return super().fill_dimensions(input_, num_dimensions)


class SimpleSelector(NeuronSelector):
    """Simply selects neurons based on a pre-defined mask."""

    def __init__(self, mask: Tensor):
        """

        Args:
            mask: the pre-defined mask for this selector

        """
        self.mask = mask

    def get_mask(self, size: List[int]) -> Tensor:
        mask_size = list(self.mask.size())
        if not mask_size == size:
            raise ValueError(f"Incorrect size. Expected: {mask_size}, got: {size}")
        return self.mask.to(midnite.get_device())


class SplitSelector(NeuronSelector):
    """Selects a number of neurons according to a specific split"""

    def __init__(self, layer_split: LayerSplit, element: List[int]):
        """
        Args:
            layer_split: the split to use
            element: the element in the split

        """
        self.layer_split = layer_split
        self.element = element

    def get_mask(self, size: List[int]) -> Tensor:
        return self.layer_split.get_mask(self.element, size)



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['midnite',
 'midnite.uncertainty',
 'midnite.visualization',
 'midnite.visualization.base',
 'midnite.visualization.base.methods']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.16,<2.0',
 'opencv-python>=4.0,<5.0',
 'pillow>=5.4,<6.0',
 'torch==1.1.0',
 'tqdm>=4.31,<5.0']

setup_kwargs = {
    'name': 'midnite',
    'version': '0.1.0',
    'description': 'This is a framework for visualization and uncertainty in CNNs.',
    'long_description': None,
    'author': 'Christina Aigner, Fabian Huch',
    'author_email': None,
    'url': 'https://luminovo.gitlab.io/midnite',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Concrete implementations of pixel activation methods."""
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import torch
import tqdm
from numpy import random
from torch import Tensor
from torch.nn import Module
from torch.optim import Adam

import midnite
from . import common
from ...base.interface import Activation
from ...base.interface import NeuronSelector
from ...base.interface import OutputRegularization
from ...base.transforms import TransformStep


class TVRegularization(OutputRegularization):
    """Total Variation norm.

    For more information, read https://arxiv.org/pdf/1412.0035v1.pdf.
    Loss is divided by number of elements to have consistent coefficients.

    """

    def __init__(self, coefficient: float = 1e2, beta: float = 2.0):
        """

        Args:
            coefficient: how much regularization to apply
            beta: beta-parameter of total variation

        """
        super().__init__(coefficient)
        self.beta = beta

    def loss(self, out: Tensor) -> Tensor:
        x_dist = (out[:, :-1, 1:] - out[:, :-1, :-1]).pow_(2)
        y_dist = (out[:, 1:, :-1] - out[:, :-1, :-1]).pow_(2)
        loss_sum = x_dist.add_(y_dist).pow_(self.beta / 2.0).sum()

        return loss_sum.div_(out.size(0) * out.size(1) * out.size(2)).mul_(
            self.coefficient
        )


class WeightDecay:
    """Weight decay regularization."""

    def __init__(self, decay_factor=1e-7):
        """

        Args:
            decay_factor: decay factor for optimizer

        """
        self.decay_factor = decay_factor


Regularization = Union[OutputRegularization, WeightDecay]


class PixelActivation(Activation):
    """Activation optimization (in Pixel space)."""

    def __init__(
        self,
        inspection_net: Module,
        top_layer_selector: NeuronSelector,
        lr: float = 1e-2,
        iter_n: int = 12,
        opt_n: int = 20,
        init_size: int = 250,
        clamp: bool = True,
        transform: Optional[TransformStep] = None,
        regularization: List[Regularization] = None,
    ):
        """
        Args:
            inspection_net: the list of adjacent layers to account for
            top_layer_selector: selector for the relevant activations
            lr: learning rate of ADAM optimizer
            iter_n: number of iterations
            opt_n: number of optimization steps per iteration
            init_size: initial width/height of visualization
            clamp: clamp image to [0, 1] space (natural image)
            transform: optional transformation after each step
            regularization: optional list of regularization terms

        """
        super().__init__([], inspection_net, top_layer_selector)
        self.lr = lr
        if iter_n < 1:
            raise ValueError("Must have at least one iteration")
        self.iter_n = iter_n
        if opt_n < 1:
            raise ValueError("Must have at least one optimization step per iteration")
        self.opt_n = opt_n
        if init_size < 1:
            raise ValueError("Initial size has to be at least one")
        self.init_size = init_size
        self.clamp = clamp
        self.transforms = transform

        if regularization is None:
            regularization = []

        # Filter out weight decay
        weight_decay = list(
            reg for reg in regularization if isinstance(reg, WeightDecay)
        )
        if len(weight_decay) > 1:
            raise ValueError("Can at most have one weight decay regularizer")
        elif len(weight_decay) == 1:
            self.weight_decay = weight_decay[0].decay_factor
        else:
            self.weight_decay = 0.0
        # Add all output regularizers
        self.output_regularizers = list(
            reg for reg in regularization if isinstance(reg, OutputRegularization)
        )

    def _opt_step(self, opt_img: Tensor) -> Tensor:
        """Performs a single optimization step.

        Args:
            opt_img: the tensor to optimize, of shape (c, h, w)

        Returns:
            the optimized image as tensor of shape (c, h, w)

        """
        # Add minibatch dim and clean up gradient
        opt_img = opt_img.unsqueeze(0).detach()

        # Optimizer for image
        optimizer = Adam([opt_img], lr=self.lr, weight_decay=self.weight_decay)

        for _ in range(self.opt_n):
            opt_img.requires_grad_(True)
            # Reset gradient
            optimizer.zero_grad()

            # Do a forward pass
            out = self.black_box_net(opt_img).squeeze(0)

            # Calculate loss (mean of output of the last layer w.r.t to our mask)
            loss = -common.calculate_single_mean(
                out, self.top_layer_selector
            ).unsqueeze(0)

            for reg in self.output_regularizers:
                loss += reg.loss(opt_img.squeeze(0))

            # Backward pass
            loss.backward()
            # Update image
            optimizer.step()

        if self.clamp:
            opt_img = opt_img.clamp(min=0, max=1)
        return opt_img.squeeze(0).detach()

    def visualize(self, input_: Optional[Tensor] = None) -> Tensor:
        # Prepare layers/input
        self.black_box_net.to(midnite.get_device())

        if input_ is None:
            # Create uniform random starting image
            input_ = torch.from_numpy(
                np.uint8(random.uniform(150, 180, (3, self.init_size, self.init_size)))
                / float(255)
            ).float()
        else:
            input_ = input_.clone().detach()
        opt_img = input_.to(midnite.get_device())

        for n in tqdm.trange(self.iter_n):
            # Optimization step
            opt_img = self._opt_step(opt_img).detach()

            # Transformation step, if necessary
            if n + 1 < self.iter_n and self.transforms is not None:
                img = opt_img.permute((1, 2, 0)).cpu().numpy()
                img = self.transforms.transform(img)
                opt_img = (
                    torch.from_numpy(img).permute(2, 0, 1).to(midnite.get_device())
                )

        return opt_img



"""Concrete implementations of backpropagation methods."""
from contextlib import contextmanager
from typing import List
from typing import Optional

from torch import Tensor
from torch.nn import functional
from torch.nn import Module
from torch.nn import Sequential

import midnite
from . import common
from ...base.interface import Attribution
from ...base.interface import LayerSplit
from ...base.interface import NeuronSelector


class Backpropagation(Attribution):
    """Propagates gradients back for a specific split."""

    def __init__(
        self,
        net: Module,
        top_layer_selector: NeuronSelector,
        bottom_layer_split: LayerSplit,
    ):
        """
        Args:
            net: the (part of the) network to backpropagate over
            top_layer_selector: relevant output neurons in the last layer
            bottom_layer_split: neuron split of the bottom layer

        """
        super().__init__([], net, top_layer_selector, bottom_layer_split)

    def visualize(self, input_: Tensor) -> Tensor:
        # Prepare layers/input
        input_ = input_.clone().detach().to(midnite.get_device()).requires_grad_()
        self.black_box_net.to(midnite.get_device())

        # forward pass
        out = self.black_box_net(input_)

        # retrieve mean score of top layer selector
        score = common.calculate_single_mean(out, self.top_layer_selector)

        # backward pass
        score.backward()

        # mean over bottom layer split dimension
        return self.bottom_layer_split.get_mean(input_.grad.detach().squeeze(dim=0))


class GuidedBackpropagation(Attribution):
    """Calculates guided gradients for a specific split."""

    def __init__(
        self,
        layers: List[Module],
        top_layer_selector: NeuronSelector,
        bottom_layer_split: LayerSplit,
    ):
        """
        Args:
            layers: the adjacent layers to perform guided gradcam on
            top_layer_selector: relevant output neurons in the last layer
            bottom_layer_split: neuron split of the bottom layer
        """
        super().__init__(layers, None, top_layer_selector, bottom_layer_split)
        self.backpropagator = Backpropagation(
            Sequential(*layers), top_layer_selector, bottom_layer_split
        )

    @staticmethod
    def _relu_grad(grad_iput: Optional[Tensor]) -> Optional[Tensor]:
        return None if grad_iput is None else functional.relu(grad_iput)

    @staticmethod
    def _gradient_relu_hook(module, grad_input, grad_output: Tensor):
        return tuple(map(GuidedBackpropagation._relu_grad, grad_input))

    @contextmanager
    def _gradient_hook_context(self):
        hooks = list(
            map(
                lambda layer: layer.register_backward_hook(self._gradient_relu_hook),
                self.white_box_layers,
            )
        )
        yield
        for hook in hooks:
            hook.remove()

    def visualize(self, input_: Tensor) -> Tensor:
        with self._gradient_hook_context():
            grad = self.backpropagator.visualize(input_)
            # Since only grad_input can be rectified via hook, apply relu for last layer
            return functional.relu(grad)



"""
Global context for controlling computation device.

While generally global state should be avoided, we believe it is correct in this
specific case - since 1) the computation device really _is_ global state of any program
running with pytorch, and 2) this state does not effect any visible branching logic.

On the contrary, making the computation device consistent avoids the frequent errors of
tensors being on different devices.

"""
import threading
from contextlib import contextmanager

import torch

# Root context is cpu device
_device_context = threading.local()
_device_context.stack = [torch.device("cpu")]


@contextmanager
def device(name: str):
    """Creates a device context for midnite to run in.

    Args:
        name: pytorch name of the device, e.g. "cpu" or "cuda:0"

    """
    _device_context.stack.append(torch.device(name))
    yield
    _device_context.stack.pop()


def get_device() -> torch.device:
    """Gets the current torch device.

    Returns: the torch device that is currently active

    """
    return _device_context.stack[-1]
